#!/usr/bin/env python3

import os
import sys
import logging
import logging.config
import argparse
from time import sleep

import schedule
import orgassist
from orgassist.config import Config, ConfigError
from orgassist.assistant import Assistant


def parse_args():
    "Parse arguments"
    p = argparse.ArgumentParser()
    p.add_argument("--config",
                   help="Path to the YAML config directory",
                   type=str,
                   default="config.yml")
    p.add_argument("--test",
                   help="Initialize, but don't start bots",
                   action="store_true",
                   default=False)
    p.add_argument("--generate-config",
                   help="Generate a new config file from a template",
                   type=str,
                   metavar="PATH",
                   default=None)

    """
    p.add_argument("--daemon",
                   help="Thread away to the background",
                   action="store_true",
                   default=False)
    """

    args = p.parse_args()
    return args


def setup_logging(cfg):
    "Configure logging"
    full_config = cfg.get('log',
                          required=False,
                          default={},
                          assert_type=dict)
    if full_config:
        full_config = dict(full_config)
        full_config['version'] = 1
        logging.config.dictConfig(full_config)
    else:
        logging.basicConfig(level=logging.WARNING)


def generate_config(cfg_path):
    "Generate a new empty config file"
    template = os.path.dirname(os.path.abspath(orgassist.__file__))
    template = os.path.join(template, 'config.tmpl.yml')
    try:
        with open(template, 'r') as handler:
            content = handler.read()
    except IOError as ex:
        print("Error while reading config template:", ex)
        return -1

    try:
        with open(cfg_path, 'w') as handler:
            handler.write(content)
    except IOError as ex:
        print("Error while reading config template:", ex)
        return -1
    return 0


def register_plugins(cfg):
    "Register external plugins specified in the config file"
    import importlib

    # Add plugin directories to the module search path
    plugins_path = cfg.get_path('plugins_path', required=False)
    if plugins_path is not None and plugins_path not in sys.path:
        sys.path.append(plugins_path)

    # Import plugins - they will register using decorators.
    plugins = cfg.get('plugins', assert_type=list)
    for plugin in plugins:
        importlib.import_module(plugin)


def setup(args):
    "Setup logging"

    cfg = Config.from_file(args.config)
    setup_logging(cfg)
    register_plugins(cfg)

    # Scheduler
    scheduler = schedule.Scheduler()

    # XMPP Bot / interface
    xmpp_bot = orgassist.bots.XmppBot(cfg.bots.xmpp)

    # Create instances of assistants
    assistants = []
    for assistant_name, assistant_config in cfg.assistants.items():

        assistant = Assistant(assistant_name,
                              assistant_config,
                              scheduler)

        assistant.register_xmpp_bot(xmpp_bot)
        # FUTURE: s.register_irc_bot(irc)

        assistants.append(assistant)

    return {
        'xmpp_bot': xmpp_bot,
        'assistants': assistants,
        'scheduler': scheduler,
    }


def main_loop(program):
    "Main loop - execute scheduled tasks"
    while True:
        program['scheduler'].run_pending()
        idle = program['scheduler'].idle_seconds
        # Sleep at most 30 seconds; bot communication works in separate
        # thread and while we are sleeping user might cause action which
        # schedules something.
        idle = min(30, idle)
        orgassist.log.debug("Scheduler sleeping %d seconds", idle)
        sleep(idle)
        program['scheduler'].run_pending()


def main():
    "Initialize environment"
    args = parse_args()
    if args.generate_config is not None:
        return generate_config(args.generate_config)

    try:
        program = setup(args)
    except ConfigError as ex:
        print("Error while parsing your configuration file:")
        print(ex.args[0])
        return 3
    except KeyboardInterrupt:
        print("Interrupted during initialization")
        return 2

    if args.test:
        # Just test
        return 0

    # Start processing in other threads
    program['xmpp_bot'].client.process()

    try:
        main_loop(program)
        # Unreachable, but keeps pylint happy
        return 5
    except KeyboardInterrupt:
        print("Keyboard interrupt")
        return 1
    finally:
        # Stop XMPP
        if program.get('xmpp_bot'):
            print("Stopping XMPP")
            program['xmpp_bot'].close()


if __name__ == "__main__":
    sys.exit(main())



"""
Glue between a calendar and other plugins and bot commands.
"""

import os
import datetime as dt
import traceback as tb
import schedule

from orgassist import log
from orgassist.config import ConfigError
from orgassist.assistant import Assistant, AssistantPlugin
from orgassist.calendar import Calendar

@Assistant.plugin('calendar')
class CalendarCore(AssistantPlugin):

    def initialize(self):
        # Scan calendar periodically and schedule notifications
        self.scheduler.every(self.scan_interval).seconds.do(self.schedule_notifications)

        # At certain points of day remind boss about agenda.
        for time in self.agenda_times:
            self.scheduler.every().day.at(time).do(self.send_agenda)

        # When scheduling notifications, store time of last scheduled event so
        # it won't be scheduled again. Do it separately for each
        self.notify_positions = {
            delta: dt.datetime.now() + dt.timedelta(minutes=delta)
            for delta in self.notify_periods
        }

    def send_notice(self, event):
        "Notify user in advance about incoming event."
        # Read just-in-time so it can be updated without restarting.
        with open(self.notice_path) as handle:
            template_content = handle.read()
        notice = event.format_notice(template_content)

        self.assistant.tell_boss(notice)

        # Do it always once only.
        return schedule.CancelJob

    def schedule_notifications(self):
        "Schedule incoming notifications"

        # Program may hang for indefinite amount of time and we still should
        # not miss any notifications. At the same time calendar may get updated,
        # tasks added or removed. We can't send duplicates.

        # Prepare incoming notifications up to 5 minutes before
        window_size = 5

        now = dt.datetime.now()

        # FIXME: EXPERIMENT, REMOVE.
        # from orgassist.calendar import Event, EventDate, DateType
        # event = Event("This is a test event. Remove me", state="TODO")
        # sched_date = EventDate(now + dt.timedelta(minutes=20), DateType.SCHEDULED)
        # event.add_date(sched_date)
        # self.scheduler.every(5).seconds.do(self.send_notice, event)

        for notify_period in self.notify_periods:
            # Calculate window
            wnd_start = now + dt.timedelta(minutes=notify_period)
            wnd_end = wnd_start + dt.timedelta(minutes=window_size)
            wnd_start = max(wnd_start, self.notify_positions[notify_period])

            last_scheduled = wnd_start
            for event in self.calendar.events:
                date = event.relevant_date.sort_date
                # We want to schedule an event if it lies between now+notify_period and
                # now+notify_period+prepare_before

                if not event.relevant_date.appointment:
                    continue

                if date <= wnd_start:
                    continue
                if date > wnd_end:
                    continue

                log.info("Scheduling %dm notification for event %r", notify_period, event)
                delta = (date - dt.datetime.now()).total_seconds() - notify_period * 60
                self.scheduler.every(delta).seconds.do(self.send_notice, event)
                last_scheduled = max(last_scheduled, date)

            self.notify_positions[notify_period] = last_scheduled


    def validate_config(self):
        "Read config and apply defaults"
        cfg = self.config
        self.notify_periods = cfg.get('notify_period',
                                      default=[5, 20])

        self.scan_interval = cfg.get('scan_interval',
                                     default=60)

        self.agenda_times = cfg.get('agenda.times',
                                    default=['7:00', '12:00'])

        #self.agenda_horizon_future = cfg.get('agenda.horizon_future',
        #                                     default=2)

        self.horizon_unfinished = cfg.get('agenda.horizon_unfinished',
                                          default=24)
        self.horizon_incoming = cfg.get('agenda.horizon_incoming',
                                        default=720)

        # Get template paths or calculate defaults
        self.agenda_path = cfg.get_path('agenda.agenda_template_path',
                                        required=False)

        self.notice_path = cfg.get_path('agenda.notice_template_path',
                                        required=False)

        def set_tmpl(filename):
            "Compute template path and check existance"
            base_path = os.path.dirname(os.path.abspath(__file__))
            base_path = os.path.join(base_path, 'templates')
            path = os.path.join(base_path, filename)
            try:
                log.debug("Trying template: %s", path)
                with open(path) as handle:
                    handle.read()
            except IOError:
                raise ConfigError('Unable to open template file: ' +
                                  path)
            return path

        self.agenda_path = set_tmpl(self.agenda_path or 'agenda.txt.j2')
        self.notice_path = set_tmpl(self.notice_path or 'notice.txt.j2')


    def register(self):
        self.calendar = Calendar(self.agenda_path)

        # Register calendar in global state - this is our public API
        self.state['calendar'] = self.calendar

        commands = [
            (['agenda', 'ag'], self.handle_agenda),
        ]
        for aliases, callback in commands:
            self.assistant.register_command(aliases, callback)

    def handle_agenda(self, message):
        "Respond with an agenda on agenda command"
        agenda = self.get_agenda()
        log.debug("Sending agenda: %r", agenda)
        message.respond(agenda)

    def get_agenda(self):
        "Generate agenda"
        now = dt.datetime.now()
        horizon_unfinished = now - dt.timedelta(hours=self.horizon_unfinished)
        horizon_incoming = now + dt.timedelta(hours=self.horizon_incoming)

        try:
            agenda = self.calendar.get_agenda(horizon_incoming,
                                              horizon_unfinished,
                                              relative_to=now)
            return agenda
        except Exception:
            tb.print_exc()
            return "Error while rendering Agenda template."

    def send_agenda(self):
        "Used for sending periodically agenda"
        agenda = self.get_agenda()
        self.assistant.tell_boss(agenda)



"""
Build configuration
"""
import setuptools

VERSION = (0, 1, 3)

with open("README.md", "r") as handler:
    LONG_DESC = handler.read()

setuptools.setup(
    name="orgassist",
    version=".".join(str(f) for f in VERSION),
    description=("Assistant who handles your appointments, tasks "
                 "and note-taking when you're away from your computer"),
    long_description=LONG_DESC,
    author="Tomasz bla Fortuna",
    author_email="bla@thera.be",
    url="https://github.com/blaa/orgassist",
    keywords="org-mode emacs bot xmpp planner",
    scripts=['assist.py'],
    include_package_data=True,
    packages=setuptools.find_packages(),
    install_requires=[
        'PyYAML==3.12',
        'dnspython==1.15.0',
        'sleekxmpp==1.3.3',
        'schedule==0.5.0',
        'Jinja2==2.10',
    ],
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Topic :: Communications :: Chat",
        "Topic :: Internet :: XMPP",
        "Topic :: Office/Business :: Scheduling",
        "Topic :: Text Editors :: Emacs",
    ],
)



"""
Abstract dates away from the org plugin so they can be used in other
modules and "agenda" handling can be shared.
"""

import datetime as dt
import enum

class DateType(enum.Enum):
    """
    Different types of events

    Smaller values sort as "more" important in relevancy algorithm.
    """
    DEADLINE = 1
    # An active date with or without timestamp
    TIMESTAMP = 2
    SCHEDULED = 3
    RANGE = 4

class EventDate:
    """
    Handle operations on a date, distinguishes different date types, date
    ranges and cyclic dates.
    """
    def __init__(self, date, date_type):
        # Date is an appointment if has hour and minutes.
        self.appointment = False

        # Handle ranges almost normally.
        if date_type == DateType.RANGE:
            assert isinstance(date, tuple)
            self.date, self.date_end = date
        else:
            self.date, self.date_end = date, None

        self.date_type = date_type
        assert date_type in DateType

        # Sortable version, always with a time.
        if not isinstance(self.date, dt.datetime):
            self.sort_date = dt.datetime(self.date.year,
                                         self.date.month,
                                         self.date.day,
                                         23, 59, 59)
        else:
            self.sort_date = self.date
            self.appointment = True

    def __lt__(self, event_date):
        "Compare dates"
        return self.sort_date < event_date.sort_date

    def is_more_relevant(self, other, relative_to=None):
        """
        Return true if self is more relevant than given date.

        relative_to defaults to "now" in the current timezone.

        Definition of the "most" relevant is in order of priority:
        - Todays appointment (accurate to minute date, today)
        - Todays date of: deadline, schedule, range start, end
        - Closest future "any" date.
        - Closest past "any" date.

        Algorithm:
        - Construct tuples:
          (-1 - future, 1 - past;
           absolute delta to now;
           date_type)
        - And use a "smaller" tuple.
        """
        if relative_to is None:
            relative_to = dt.datetime.now()

        # Future is positive
        delta_this = (self.sort_date - relative_to).total_seconds()
        delta_other = (other.sort_date - relative_to).total_seconds()

        this_tuple = (
            -1 if delta_this >= 0 else 1,
            abs(delta_this),
            self.date_type.value
        )

        other_tuple = (
            -1 if delta_other >= 0 else 1,
            abs(delta_other),
            other.date_type.value
        )

        # TODO: Remove when certain, the algorithm is ok.
        """
        print("\nCOMPARE")
        print("  ", self, other)
        print("  ", delta_this, delta_other)
        print("  ", this_tuple, other_tuple)
        if this_tuple <= other_tuple:
            print("  THIS")
        else:
            print("  OTHER")
        """
        return this_tuple <= other_tuple

    def __repr__(self):
        "Display EventDates nicely"
        def format_date(date):
            "Short format for date"
            if isinstance(date, dt.datetime):
                return date.strftime("%Y-%m-%d %H:%M")
            return date.strftime("%Y-%m-%d")

        if self.date_type == DateType.RANGE:
            ranged = '<->[%s]' % format_date(self.date_end)
        else:
            ranged = ''
        txt = '<EventDate [%s]%s %s>' % (format_date(self.date), ranged,
                                         self.date_type.name)
        return txt



"""
Abstract events away from the org plugin so they can be used in other
modules and "agenda" handling can be shared.
"""
import datetime as dt
import jinja2

class EventState:
    """
    Different event sources can have different states. They are normally a
    short string which doesn't mean much. For eg. this can be: TODO, DONE,
    DELEGATED, CANCELLED.

    Some of those are considered DONE, other belong in an "Open" group and
    shoukd be listed in agendas.
    """
    def __init__(self, name, is_open=None):
        self.name = name
        if is_open is None:
            self.is_open = EventState.is_state_open(name)
        else:
            self.is_open = is_open

    @staticmethod
    def is_state_open(state):
        "Handle default states"
        open_map = {
            'TODO': True,
            'DELEGATED': True,
            'BLOCKED': True,
            'DONE': False,
            'CANCELLED': False,
        }
        try:
            return open_map[state]
        except KeyError:
            raise Exception("State %s has unknown default open/close state" % state)

    def __repr__(self):
        return '<EventState %s open=%s>' % (self.name,
                                            self.is_open)

class Event:
    """
    Abstracts a calendar event from plugins.
    """
    def __init__(self, headline, state=None):
        """Initialize event variables"""

        self.headline = headline

        # Set state
        self.set_state(state)

        self.tags = set()

        # A, B, C (letter)
        self.priority = None

        # The "most relevant date *now*".
        # 1) The next future day of the event (for example of a cyclic event)
        # 2) Or the closest "todays" date
        # 3) Last past date for unfinished/past events.
        self.relevant_date = None

        # Event can have multiple dates of various types.
        self.dates = []

        # Set of all date types for this event
        self.date_types = set()

        # Event body content
        self.body = ""

        # Metadata, eg. location in org-mode tree.
        # Controlled freely by creator.
        self.meta = {}

    def add_date(self, event_date, relative_to=None):
        "Add date to the event"
        assert event_date not in self.dates

        self.dates.append(event_date)

        # Update relevant date
        if self.relevant_date is None:
            self.relevant_date = event_date
        elif event_date.is_more_relevant(self.relevant_date,
                                         relative_to=relative_to):
            self.relevant_date = event_date

        # Update event type
        self.date_types.add(event_date.date_type)

    def set_state(self, state):
        "Handle state as object or string - if string, cast to object"
        if isinstance(state, str):
            self.state = EventState(state)
        else:
            self.state = state

    def add_tags(self, tags):
        "Add tags to the event"
        if isinstance(tags, str):
            tags = {tags}
        self.tags.update(tags)
        return self

    def format_notice(self, template_content):
        """
        Render a notice about this event using given template.

        If you are authoring a plugin which has weird additional metadata you
        can override this method.
        """
        # Jinja context
        now = dt.datetime.now()

        # Calculate time delta in conscise human-readable way
        delta_s = (self.relevant_date.sort_date - now).total_seconds()

        delta_h = delta_s // 60 // 60
        delta_s -= delta_h * 60 * 60

        delta_m = delta_s // 60
        delta_s -= delta_m * 60

        # If scheduler gets late 10 seconds - round the number.
        if delta_s > 50:
            delta_m += 1
            delta_s = 0

        delta_str = '%dh' % delta_h if delta_h > 0 else ''
        delta_str += '%dm' % delta_m if delta_m > 0 or delta_h > 0 else ''

        ctx = {
            'now': now,
            'event': self,
            'delta_str': delta_str,

            # Shortcuts
            'headline': self.headline,
            'relevant_date': self.relevant_date,
            'date': self.relevant_date.sort_date,
            # To differentiate template based on event subclasses
            'event_cls': self.__class__.__name__,
        }

        if template_content is None:
            return ctx

        template = jinja2.Template(template_content)
        rendered = template.render(ctx)

        return rendered

    def __repr__(self):
        "Useful debugging representation"
        add_up = [
            field if isinstance(field, str) else repr(field)
            for field in [
                "'" + self.headline[:20] + "'",
                self.priority,
                ('state=' + self.state.name) if self.state is not None else '',
                self.relevant_date,
                ','.join(self.tags),
            ]
            if field is not None
        ]
        return "<Event %s>" % " ".join(add_up)

    def __lt__(self, other):
        return self.relevant_date < other.relevant_date



"""
Console based reporting for data returned by orghelpers.

(C) Tomasz bla Fortuna
"""
import datetime as dt

def _get_marker(eventtype):
    if eventtype == 'DEADLINE':
        marker = 'D'
    elif eventtype == 'SCHEDULED':
        marker = 'S'
    elif eventtype == 'RANGE':
        marker = 'R'
    else:
        marker = ' '
    return marker


def _get_delta(delta, accurate=False):
    import math
    days = int(math.floor(delta))
    if days == 0:
        # Today
        if accurate:
            hours = math.floor(delta*24)
            if hours <= 0.1:
                return "NOW"
            if hours == 1:
                return "today in " + str(int(hours)) + " hour"
            if hours > 1:
                return "today in " + str(int(hours)) + " hours"
        else:
            return "today"
    elif days > 1:
        return "in " + str(days) + " days"
    elif days < 0:
        return str(-days) + " days ago"
    elif days == 1:
        return "1 day"


def report_stat(incoming_list, tasks_open, tasks_all):
    "Create a simple statistic for following days"
    incoming_list.sort()

    now = dt.datetime.now()
    eo_today = dt.datetime(now.year, now.month, now.day, 23, 59, 59)
    eo_tomorrow = eo_today + dt.timedelta(days=1)

    counted = set()

    stat_today = 0
    stat_tomorrow = 0
    stat_total = 0

    for incoming in incoming_list:
        closest_converted_date, data = incoming
        if data['entry'] in counted:
            continue # Count once, earliest entry
        counted.add(data['entry'])

        if closest_converted_date <= eo_today:
            stat_today += 1
        elif closest_converted_date <= eo_tomorrow:
            stat_tomorrow += 1

        stat_total += 1

    stat_rest = stat_total - stat_today - stat_tomorrow

    s = ""
    if stat_today:
        s += str(stat_today)
    if stat_tomorrow:
        s += '->%d' % stat_tomorrow
    if stat_rest:
        s += '-->%d' % stat_rest

    #s = "T %d->%d-->%d" % (stat_today, stat_tomorrow, stat_rest)

    # Not timed
    s += " %d/%d" % (tasks_open, tasks_all)
    return s


def report_incoming(incoming_list, cfg):
    "Report incoming tasks for following days"
    if not incoming_list:
        return

    today = dt.datetime.today()
    incoming_list.sort()
    add_separator = False # Today separator

    if incoming_list[0][0].day == today.day:
        add_separator = True

    for incoming in incoming_list:
        closest_converted_date, data = incoming
        todo = data['entry'].todo or 'TASK'

        if add_separator and closest_converted_date.day != today.day:
            add_separator = False
            print("- EOD -")

        marker = _get_marker(data['eventtype'])
        accurate = data['converted_date'] == data['date']
        # If equal - then the event has a time specified.
        if accurate and data['delta']*24 < cfg['mark_in']:
            marker += '--> '
        else:
            marker += '    '

        s = "%s %9s %-20s %s"
        s = s % (marker, todo, _get_delta(data['delta'], accurate),
                 data['entry'].headline[:60])
        print(s)


def report_unfinished(unfinished_list, cfg):
    "Report incoming tasks for following days"
    if not unfinished_list:
        return

    unfinished_list.sort()
    output = False

    for unfinished in unfinished_list:
        closest_converted_date, data = unfinished
        todo = data['entry'].todo

        if data['eventtype'] not in ['SCHEDULED', 'DEADLINE']:
            continue # Don't show things with plain timestamps

        marker = _get_marker(data['eventtype'])
        marker += ' DUE'
        accurate = data['converted_date'] == data['date']

        s = "%s %9s %-20s %s"
        s = s % (marker, todo, _get_delta(data['delta'], accurate),
                 data['entry'].headline[:60])

        output = True
        print(s)
    return output


def report_projects(projects):
    "Report an open projects you're responsible for"
    if not projects:
        return

    print("PROJECTS:")
    for project, stat in projects.items():
        print("  %-30s" % project.headline[:50], end=' ')
        stats = " ".join(["%s=%d" % (k, v) for k, v in stat.items()])
        if stats:
            print("(%s)" % stats)
        else:
            print()
    print()
